"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const as_mutable_1 = require("as-mutable");
const is_1 = require("../utils/is");
const action_1 = require("./action");
const reducerFactory_1 = require("./reducerFactory");
// interface IConfigObject<S, P, R, E> {
//   namespace?: string;
//   path?: string;
//   pathParams?: P;
//   defaultState?: S;
//   methods: R;
// }
// interface IService {
//   namespace?: string;
//   path?: string;
//   pathParams?: any;
//   reducer: any;
//   actions: any;
// }
function getPropValueFrom(targets, prop, defaultValue) {
    const value = targets.reduce((val, target) => (val === undefined && prop in target ? target[prop] : val), undefined);
    return value === undefined ? defaultValue : value;
}
// for OOP mode
const getContextFactory = proto => (state, path, { stateCache }) => {
    // get current state from cache
    if (stateCache && stateCache.has && stateCache.has(path)) {
        state = stateCache.get(path);
    }
    const self = {
        // private API
        __mutableState__: undefined,
        __getState__() {
            return as_mutable_1.getValue(this.__mutableState__);
        },
        // public API
        get state() {
            return this.__mutableState__;
        },
        setState(partialState) {
            Object.assign(this.__mutableState__, partialState);
        }
    };
    self.__proto__ = proto;
    // when invoke `getContext(state)`, initialize `__mutableState__`
    self.__mutableState__ = as_mutable_1.asMutable(state);
    // cache current state
    if (stateCache && stateCache.has && !stateCache.has(path)) {
        stateCache.set(path, self.__mutableState__);
    }
    return self;
};
function produceDeriveFuncFromGetters(getters) {
    let derive;
    const getterKeys = Object.keys(getters);
    if (getterKeys.length) {
        derive = function (state) {
            if (is_1.isPlainObject(state)) {
                const draftState = Object.assign({}, state);
                Object.defineProperties(draftState, getters);
                getterKeys.forEach(key => {
                    const value = draftState[key];
                    Object.defineProperty(state, 'key', { value, writable: true });
                });
                return state;
            }
            else {
                return state;
            }
        };
    }
    return derive;
}
function createService(config, override = {}) {
    const nsp = getPropValueFrom([override, config], 'namespace', '');
    const path = getPropValueFrom([override, config], 'path', '').replace(/\/{1,}$/, ''); // trim trailing slashes
    const params = getPropValueFrom([override, config], 'pathParams');
    const defaultState = getPropValueFrom([override, config], 'defaultState');
    // TODO: some validations on config props
    const methods = {};
    let actions = {};
    let derive;
    let getContext;
    // assign `methods` and `derive`
    if (is_1.isFunction(config)) {
        const proto = config.prototype || {};
        getContext = getContextFactory(proto);
        const getters = {};
        Object.getOwnPropertyNames(proto).forEach(key => {
            if (key === 'constructor')
                return;
            const desc = Object.getOwnPropertyDescriptor(proto, key) || {};
            if (is_1.isFunction(desc.value)) {
                const method = desc.value;
                // convert OOP method signature to standard reducer signature
                methods[key] = function (state, action) {
                    const ret = method.call(this, ...action.payload);
                    if (is_1.isIterator(ret))
                        return ret;
                    return this.__getState__();
                };
            }
            else if (desc.get) {
                getters[key] = desc;
            }
        });
        derive = produceDeriveFuncFromGetters(getters);
    }
    else {
        // mode = 'FP';
        const configMethods = Object.assign({}, config.methods);
        getContext = () => (Object.assign({}, configMethods));
        Object.keys(configMethods).forEach(key => {
            const method = configMethods[key];
            // convert FP method signature to standard reducer signature
            methods[key] = function (state, action) {
                return method.call(this, state, ...action.payload);
            };
        });
        derive = config.derive;
    }
    // assign `actions`
    for (const methodName in methods) {
        actions[methodName] = (...args) => {
            return {
                type: action_1.createActionType(nsp, methodName, path),
                payload: args
            };
        };
    }
    if (params) {
        actions = action_1.withPathParams(actions);
    }
    const reducer = reducerFactory_1.reducerFactory({
        nsp,
        path,
        params,
        methods,
        getContext,
        derive,
        defaultState
    });
    // extra prop to inform gateway about self path
    if (params) {
        reducer.params = params;
    }
    else {
        reducer.path = path;
    }
    Object.assign(reducer, { asService: true, reducer, actions });
    return reducer;
}
exports.createService = createService;
