"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("./path");
const task_1 = require("./task");
const getState_1 = require("./getState");
const action_1 = require("./action");
const is_1 = require("../utils/is");
const finalizeReducer = (state, action) => {
    if (action.type === '@@UPDATE' && action.meta && action.meta.finalize) {
        const localState = path_1.getByPath(state, action.meta.path);
        const updater = action.payload;
        const nextState = updater(localState);
        return path_1.setByPath(state, action.meta.path, nextState);
    }
    else {
        return state;
    }
};
class ServiceGateway {
    constructor() {
        this.store = {};
        this.reducers = [];
        this.taskRunner = {};
        // instantiated enhancer:
        this.enhancer = createStore => (reducer, initialState, enhancer) => {
            this.store.taskRunner = this.taskRunner = new task_1.TaskRunner(this.store);
            this.store.stateCache = new Map();
            const _store = createStore(reducer, initialState, enhancer);
            Object.assign(this.store, _store);
            this.store.getState = getState_1.getStateFactory(_store.getState);
            this.store.dispatch = action => {
                const value = _store.dispatch(action);
                this.taskRunner.run();
                return value;
            };
            return this.store;
        };
        this.reducer = (rootState, action, ...args) => {
            rootState = finalizeReducer(rootState, action);
            return this.reducers.reduce((rootState, reducer) => {
                if (reducer.asService) {
                    let effectivePath;
                    if (reducer.params) {
                        const parsed = action_1.parseAction(action);
                        effectivePath = parsed && parsed.path;
                    }
                    else {
                        effectivePath = typeof reducer.path === 'string' ? reducer.path : '';
                    }
                    if (typeof effectivePath !== 'string')
                        return rootState;
                    const localState = path_1.getByPath(rootState, effectivePath);
                    const nextState = reducer(localState, action, this.store);
                    if (is_1.isIterator(nextState)) {
                        this.taskRunner.registerTask({ path: effectivePath, work: nextState });
                        return rootState;
                    }
                    else {
                        this.store.stateCache.delete(effectivePath);
                        return path_1.setByPath(rootState, effectivePath, nextState);
                    }
                }
                else {
                    // handle normal reducer
                    return args.length ? reducer(rootState, action, ...args) : reducer(rootState, action);
                }
            }, rootState);
        };
        this.use = (...reducers) => {
            reducers.forEach(reducer => {
                if (this.reducers.indexOf(reducer) !== -1)
                    return false;
                this.reducers = this.reducers.concat(reducer);
                this.reducers.sort((r1, r2) => path_1.comparePath(r1.path, r2.path));
                return true;
            });
        };
        this.unuse = reducer => {
            if (this.reducers.indexOf(reducer) === -1)
                return false;
            this.reducers = this.reducers.filter(r => r !== reducer);
            return true;
        };
    }
}
exports.ServiceGateway = ServiceGateway;
exports.createServiceGateway = () => {
    return new ServiceGateway();
};
